DEFAULT_YOLO_INPUT_SIZE = 416
DEFAULT_YOLO_IOU = 0.45
DEFAULT_YOLO_FRAMEWORK = 'tf'
DEFAULT_YOLO_SCORE_THRESHOLD = 0.50
DEFAULT_YOLO_TINY = False
DEFAULT_YOLO_DOWNLOAD_WEIGHT_PATH = '/etc/image_detection-service/models/yolo/yolov4.weights'
DEFAULT_YOLO_FINAL_WEIGHT_PATH = '/etc/image_detection-service/models/yolo_checkpoint'
DEFAULT_YOLO_MODEL_NAME = "yolov4"


DEFAULT_YOLO_CLASS_FILE = '/etc/example_templete/data/classes/coco.names'
DEFAULT_YOLO_ANNOTATION_FILE = '/etc/example_templete/data/dataset/val2017.txt'
DEFAULT_YOLO_ANNOTATION_FILE_TEST = "/etc/example_templete/data/dataset/val2017.txt"
DEFAULT_YOLO_DETECTION_PATH = "/etc/example_templete/data/detection/"
